import tomllib
from dataclasses import dataclass
from pathlib import Path
from typing import Literal

from spidr.config import DataConfig, MaskingConfig, OptimizerConfig, RunConfig


@dataclass(frozen=True)
class UserConfig:
    num_classes: int
    manifest: str
    workdir: str
    wandb_project: str
    wandb_name: str
    wandb_mode: Literal["online", "offline"]


def hubert_optim_config() -> OptimizerConfig:
    return OptimizerConfig(
        betas=(0.9, 0.98),
        init_lr_scale=1e-8,
        final_lr_scale=1e-8,
        warmup_steps=32_000,
        hold_steps=0,
        decay_steps=368_000,
        to_freeze=[],
    )


def hubert_data_config(manifest: str) -> DataConfig:
    return DataConfig(
        manifest,
        normalize=True,
        min_sample_size=32_000,
        max_sample_size=250_000,
        max_batch_length=2_800_000,
        num_buckets=100,
        num_workers=24,
        prefetch_factor=4,
        bucket_method="percentile",
    )


@dataclass(frozen=True)
class Config:
    """Full configuration."""

    num_classes: int
    run: RunConfig
    data: DataConfig
    optimizer: OptimizerConfig
    masking: MaskingConfig


def read_config(path: str | Path) -> Config:
    user = UserConfig(**tomllib.loads(Path(path).read_text(encoding="utf-8")))
    return Config(
        num_classes=user.num_classes,
        run=RunConfig(user.workdir, user.wandb_project, user.wandb_name, user.wandb_mode),
        data=hubert_data_config(user.manifest),
        optimizer=hubert_optim_config(),
        masking=MaskingConfig(),
    )



import pytest
import torch
from s3prl.upstream.hubert.hubert_model import HubertConfig, HubertModel, HubertPretrainingConfig

from minimal_hubert import HuBERTPretrain as MyHuBERT
from minimal_hubert.model import state_dict_from_s3prl


@pytest.fixture(scope="module")
def num_classes() -> int:
    return 100


@pytest.fixture(scope="module")
def s3prl_hubert(num_classes: int, device: torch.device) -> HubertModel:
    cfg = HubertConfig(label_rate=50, final_dim=256)
    return HubertModel(cfg, HubertPretrainingConfig(), dictionaries=[range(num_classes)]).eval().to(device)


@pytest.fixture(scope="module")
def my_hubert(s3prl_hubert: HubertModel, num_classes: int, device: torch.device) -> MyHuBERT:
    model = MyHuBERT(num_classes).eval().to(device)
    model.load_state_dict(state_dict_from_s3prl(s3prl_hubert.state_dict()))
    return model


@pytest.mark.filterwarnings("ignore::FutureWarning")
@torch.no_grad
def test_s3prl_encoder_intermediate(my_hubert: MyHuBERT, s3prl_hubert: HubertModel, waveforms: torch.Tensor) -> None:
    y = my_hubert.get_intermediate_outputs(waveforms, before_residual=False)
    for layer, yi in enumerate(y):
        xi = s3prl_hubert.extract_features(waveforms, output_layer=layer + 1)[0]
        torch.testing.assert_close(xi, yi)



"""Entry point for launching pretraining."""

import argparse
from pathlib import Path

from spidr.slurm import launch_with_submitit, slurm_config_parse_args
from spidr.tools import init_logger

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Launch HuBERT pretraining.", allow_abbrev=False)
    parser.add_argument("configs", type=Path, nargs="+", help="TOML config file(s) for the training run(s).")
    parser.add_argument("-A", "--account", type=str, required=True, help="Slurm account")
    parser.add_argument("-N", "--nodes", type=int, required=True, help="Number of nodes")
    parser.add_argument("-G", "--gpus-per-node", type=int, required=True, help="GPUs per node")
    parser.add_argument("-c", "--cpus-per-task", type=int, help="CPUs per task")
    parser.add_argument("--mem-per-gpu", type=str, help="Memory per GPU")
    parser.add_argument("-t", "--time", type=int, help="Time limit in minutes")
    parser.add_argument("-C", "--constraint", type=str, help="Slurm constraint")
    parser.add_argument("-p", "--partition", type=str, help="Slurm partition")
    parser.add_argument("-q", "--qos", type=str, help="Slurm QoS")
    parser.add_argument("--dump", type=Path, help="Submitit dump", required=True)
    args = parser.parse_args()

    from .config import read_config
    from .train import train

    init_logger()
    jobs = [(train, (read_config(cfg),)) for cfg in args.configs]
    name = jobs[0][1][0].run.wandb_name
    launch_with_submitit(name, jobs, args.dump, slurm_config_parse_args(args), copy_code=False)



import pytest
import torch
from torch.testing import make_tensor


@pytest.fixture(scope="session")
def device() -> torch.device:
    return torch.device("cuda" if torch.cuda.is_available() else "cpu")


@pytest.fixture
def waveforms(device: torch.device) -> torch.Tensor:
    return make_tensor((8, 16_000), dtype=torch.float32, low=0, high=10, device=device)



import pytest
import torch
from transformers import HubertConfig, HubertModel

from minimal_hubert import HuBERT as MyHuBERT
from minimal_hubert.model import state_dict_from_torchaudio_or_huggingface


@pytest.fixture(scope="module")
def hf_hubert(device: torch.device) -> HubertModel:
    return HubertModel(HubertConfig()).eval().to(device)


@pytest.fixture(scope="module")
def my_hubert(hf_hubert: HubertModel, device: torch.device) -> MyHuBERT:
    model = MyHuBERT().eval().to(device)
    model.load_state_dict(state_dict_from_torchaudio_or_huggingface(hf_hubert.state_dict()))
    return model


@torch.no_grad
def test_hf_encoder_intermediate(my_hubert: MyHuBERT, hf_hubert: HubertModel, waveforms: torch.Tensor) -> None:
    x = hf_hubert(waveforms, output_hidden_states=True)
    y = my_hubert.get_intermediate_outputs(waveforms, before_residual=False)
    torch.testing.assert_close(x.last_hidden_state, y[-1])
    for xi, yi in zip(x.hidden_states[1:], y, strict=True):
        torch.testing.assert_close(xi, yi)



import math
import os
import socket
from collections.abc import Sequence
from pathlib import Path


def split_for_distributed[T](sequence: Sequence[T]) -> Sequence[T]:
    if "SLURM_NTASKS" not in os.environ:
        return sequence
    rank, world_size = int(os.environ["SLURM_PROCID"]), int(os.environ["SLURM_NTASKS"])
    array_id, num_arrays = int(os.getenv("SLURM_ARRAY_TASK_ID", "0")), int(os.getenv("SLURM_ARRAY_TASK_COUNT", "1"))
    if "SLURM_ARRAY_TASK_ID" in os.environ:
        assert os.environ["SLURM_ARRAY_TASK_MIN"] == "0"
        assert int(os.environ["SLURM_ARRAY_TASK_MAX"]) == num_arrays - 1

    n_total = len(sequence)  # Split by array first
    files_per_array = math.ceil(n_total / num_arrays)
    start = array_id * files_per_array
    end = min(start + files_per_array, n_total)
    sequence = sequence[start:end]

    n_local = len(sequence)  # Then split by rank within each array
    files_per_rank = math.ceil(n_local / world_size)
    start = rank * files_per_rank
    end = min(start + files_per_rank, n_local)
    return sequence[start:end]


def slurm_job_tmpdir() -> Path | None:
    if "JOBSCRATCH" in os.environ:
        return Path(os.environ["JOBSCRATCH"])
    if (path := Path(f"/fastscratch/{socket.gethostname()}")).is_dir() and "SLURM_JOB_ID" in os.environ:
        return path / os.environ["SLURM_JOB_ID"]
    return None



from pathlib import Path

import orjson
import torch
from filelock import FileLock
from sklearn.cluster import MiniBatchKMeans
from tqdm import tqdm

from .features import split_for_distributed


def transcribe(root: str | Path, kmeans: MiniBatchKMeans, jsonl: str | Path) -> None:
    root = Path(root)
    files = split_for_distributed(sorted(root.rglob("*.pt")))
    lock = FileLock(f"{jsonl}.lock")
    for path in tqdm(files):
        fileid = str(path.relative_to(root)).removesuffix(".pt")
        units = kmeans.predict(torch.load(path).numpy()).tolist()
        units_str = orjson.dumps({"file": fileid, "units": units}, option=orjson.OPT_APPEND_NEWLINE)
        with lock, Path(jsonl).open("ab") as f:
            f.write(units_str)


if __name__ == "__main__":
    import argparse

    import joblib

    parser = argparse.ArgumentParser(description="Inference of discrete units")
    parser.add_argument("root_features", type=Path, help="Root directory containing feature files")
    parser.add_argument("path_kmeans", type=Path, help="Path to the trained K-means model")
    parser.add_argument("output_jsonl", type=Path, help="Path to the output JSONL file with units")
    args = parser.parse_args()
    transcribe(args.root_features, joblib.load(args.path_kmeans), args.output_jsonl)



from os.path import commonpath
from pathlib import Path

import torch
from spidr.data import speech_dataset
from torch import Tensor
from torchaudio.compliance import kaldi
from torchaudio.functional import compute_deltas
from tqdm import tqdm

from .model import HuBERTPretrain
from .utils import split_for_distributed


def mfcc(waveform: Tensor) -> Tensor:
    mfccs = kaldi.mfcc(waveform).transpose(0, 1)
    deltas = compute_deltas(mfccs)
    ddeltas = compute_deltas(deltas)
    return torch.cat([mfccs, deltas, ddeltas], dim=0).transpose(0, 1).contiguous()


def compute_and_save_mfccs(path_manifest: str | Path, root_features: str | Path) -> None:
    dataset = speech_dataset(path_manifest, normalize=False)
    root = commonpath(dataset.manifest["path"])
    indices = split_for_distributed(list(range(len(dataset))))
    dest = Path(root_features)
    for i in tqdm(indices):
        name = Path(dataset.manifest[i, "path"]).relative_to(root).with_suffix(".pt")
        waveform = dataset[i].unsqueeze(0)
        features = mfcc(waveform)
        (dest / name.parent).mkdir(exist_ok=True, parents=True)
        torch.save(features, dest / name)


@torch.no_grad()
def compute_and_save_hubert_features(
    path_manifest: str | Path,
    root_features: str | Path,
    checkpoint: str | Path,
    layer: int,
) -> None:
    dataset = speech_dataset(path_manifest, normalize=True)
    root = commonpath(dataset.manifest["path"].to_list())
    indices = split_for_distributed(list(range(len(dataset))))
    dest = Path(root_features)
    model = HuBERTPretrain.from_pretrained(checkpoint).cuda()
    for i in tqdm(indices):
        name = Path(dataset.manifest[i, "path"]).relative_to(root).with_suffix(".pt")
        waveform = dataset[i].unsqueeze(0).cuda()
        features = model.get_intermediate_outputs(waveform, num_layers=layer)[-1].squeeze().cpu()
        (dest / name.parent).mkdir(exist_ok=True, parents=True)
        torch.save(features, dest / name)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Extract features using MFCC or HuBERT")
    parser.add_argument("manifest", type=Path, help="Path to the manifest file")
    parser.add_argument("features", type=Path, help="Root directory for output features")
    parser.add_argument("--type", required=True, choices=["mfcc", "hubert"], help="Feature type: 'mfcc' or 'hubert'")
    parser.add_argument("--checkpoint", type=Path, help="Path to HuBERT checkpoint file")
    parser.add_argument("--layer", type=int, help="Layer number to extract features from")
    args = parser.parse_args()
    if args.type == "mfcc":
        compute_and_save_mfccs(args.manifest, args.features)
    else:
        compute_and_save_hubert_features(args.manifest, args.features, args.checkpoint, args.layer)



from collections.abc import Iterable
from pathlib import Path
from tempfile import TemporaryDirectory

import orjson
import torch
from fastabx import zerospeech_abx
from filelock import FileLock
from spidr.config import SAMPLE_RATE
from torch.nn import functional as F
from torchcodec.decoders import AudioDecoder
from tqdm import tqdm

from .model import HuBERTPretrain
from .utils import slurm_job_tmpdir


def extract_features(
    model: HuBERTPretrain,
    audio: Path,
    features: Path,
    layers: set[int],
    *,
    extension: str,
) -> None:
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model = model.to(device)
    num_layers = max(layers)
    for path in tqdm(sorted(audio.rglob(f"*{extension}")), desc="Extract features"):
        samples = AudioDecoder(path).get_all_samples()
        assert samples.sample_rate == SAMPLE_RATE
        data = F.layer_norm(samples.data.to(device), samples.data.shape)
        for i, feats in enumerate(model.get_intermediate_outputs(data, num_layers=num_layers)):
            if i + 1 not in layers:
                continue
            dest = features / str(i + 1) / path.with_suffix(".pt").name
            dest.parent.mkdir(exist_ok=True, parents=True)
            torch.save(feats.squeeze().cpu(), dest)


def compute_and_save_abx(
    item: str | Path,
    audio: str | Path,
    output: str | Path,
    *,
    checkpoint: str | Path,
    layers: Iterable[int] | None = None,
    extension: str = ".wav",
    tmpdir: str | Path | None = None,
) -> None:
    item_name = Path(item).stem
    model = HuBERTPretrain.from_pretrained(checkpoint)
    layers = set(range(1, 13) if layers is None else layers)
    assert min(layers) >= 1
    assert max(layers) <= 12
    lock = FileLock(f"{output}.lock")
    with TemporaryDirectory(dir=tmpdir) as features:
        extract_features(model, Path(audio), Path(features), layers, extension=extension)
        for layer in layers:
            this_features = Path(features) / str(layer)
            for speaker in ("within", "across"):
                score = zerospeech_abx(item, this_features, max_size_group=10, max_x_across=5, speaker=speaker)
                result = {"item": item_name, "layer": layer, "speaker": speaker, "score": score}
                with lock, Path(output).open("ab") as f:
                    f.write(orjson.dumps(result, option=orjson.OPT_APPEND_NEWLINE))


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Compute ABX discriminability of HuBERT representations")
    parser.add_argument("item", type=Path, help="Path to the item file")
    parser.add_argument("audio", type=Path, help="Path to the audio files")
    parser.add_argument("output", type=Path, help="Path to the output JSONL file with ABX scores")
    parser.add_argument("--checkpoint", required=True, help="Path to the HuBERT checkpoint file")
    parser.add_argument("--layers", nargs="+", type=int, help="Target layers")
    parser.add_argument("--extension", type=str, default=".wav", help="Audio file extension")
    parser.add_argument(
        "--tmpdir",
        type=Path,
        help="Path to the temporary directory where features will be stored",
        default=slurm_job_tmpdir(),
    )
    args = parser.parse_args()

    compute_and_save_abx(
        args.item,
        args.audio,
        args.output,
        checkpoint=args.checkpoint,
        layers=args.layers,
        extension=args.extension,
        tmpdir=args.tmpdir,
    )



from copy import deepcopy

import pytest
import torch
from spidr.data.dataset import DEFAULT_CONV_LAYER_CONFIG, conv_length
from torch import Tensor, nn
from torch.nn import functional as F
from torchaudio.models import HuBERTPretrainModel as ReferenceHuBERT
from torchaudio.models import hubert_pretrain_base

from minimal_hubert import HuBERTPretrain as MyHuBERT
from minimal_hubert.model import state_dict_from_torchaudio_or_huggingface


@pytest.fixture(scope="module")
def torchaudio_hubert(device: torch.device) -> ReferenceHuBERT:
    return hubert_pretrain_base().eval().to(device)


@pytest.fixture(scope="module")
def num_classes(torchaudio_hubert: ReferenceHuBERT) -> int:
    return torchaudio_hubert.logit_generator.label_embeddings.size(0)


@pytest.fixture(scope="module")
def my_hubert(torchaudio_hubert: ReferenceHuBERT, num_classes: int, device: torch.device) -> MyHuBERT:
    model = MyHuBERT(num_classes).eval().to(device)
    model.load_state_dict(state_dict_from_torchaudio_or_huggingface(torchaudio_hubert.state_dict()))
    return model


@torch.no_grad
def test_torchaudio_encoder_forward(
    my_hubert: MyHuBERT,
    torchaudio_hubert: ReferenceHuBERT,
    waveforms: Tensor,
) -> None:
    x, _ = torchaudio_hubert.wav2vec2(waveforms)
    y = my_hubert.feature_extractor(waveforms)
    y = my_hubert.feature_projection(y)
    y = my_hubert.encoder(y)
    torch.testing.assert_close(x, y)


@torch.no_grad
def test_torchaudio_encoder_intermediate(
    my_hubert: MyHuBERT,
    torchaudio_hubert: ReferenceHuBERT,
    waveforms: Tensor,
) -> None:
    x, _ = torchaudio_hubert.wav2vec2.extract_features(waveforms)
    y = my_hubert.get_intermediate_outputs(waveforms, before_residual=False)
    for xi, yi in zip(x, y, strict=True):
        torch.testing.assert_close(xi, yi)


class DummyMaskGenerator(nn.Module):
    def __init__(self, embedding: nn.Parameter) -> None:
        super().__init__()
        self.mask_embedding = embedding

    def get_mask(self, shape: tuple[int, ...]) -> Tensor:
        torch.manual_seed(0)
        proba = 0.5
        return torch.randn(shape, dtype=torch.float32) > proba

    def forward(self, x: Tensor, _: Tensor | None = None) -> tuple[Tensor, Tensor | None]:
        mask = self.get_mask((x.size(0), x.size(1))).to(x.device)
        x[mask] = self.mask_embedding.to(x.dtype)
        return x, mask


def torchaudio_hubert_loss(
    logit_m: Tensor,
    logit_u: Tensor | None,
    feature_penalty: Tensor,
    masked_weight: float,
    unmasked_weight: float,
    feature_weight: float,
    reduction: str,
) -> Tensor:
    """From https://github.com/pytorch/audio/blob/main/examples/hubert/loss/hubert_loss.py"""
    loss = feature_penalty * feature_weight * logit_m.shape[0]
    target_m = torch.zeros(logit_m.shape[0], dtype=torch.long, device=logit_m.device)
    loss_m = F.cross_entropy(logit_m, target_m, reduction=reduction)
    loss += loss_m * masked_weight
    if logit_u is not None:
        target_u = torch.zeros(logit_u.shape[0], dtype=torch.long, device=logit_m.device)
        loss_u = F.cross_entropy(logit_u, target_u, reduction=reduction)
        loss += loss_u * unmasked_weight
    return loss


@torch.no_grad
def test_torchaudio_loss(
    my_hubert: MyHuBERT,
    torchaudio_hubert: ReferenceHuBERT,
    waveforms: Tensor,
    num_classes: int,
) -> None:
    torchaudio_hubert = deepcopy(torchaudio_hubert)
    mask_generator = DummyMaskGenerator(torchaudio_hubert.mask_generator.mask_embedding)
    torchaudio_hubert.mask_generator = mask_generator

    shape = (waveforms.size(0), int(conv_length(DEFAULT_CONV_LAYER_CONFIG, waveforms.size(1))))
    labels = torch.randint(0, num_classes, shape, device=waveforms.device)
    mask = mask_generator.get_mask(shape).to(waveforms.device)
    x = torchaudio_hubert_loss(
        *torchaudio_hubert(waveforms, labels),
        masked_weight=1.0,
        unmasked_weight=0.0,
        feature_weight=1.0,
        reduction="mean",
    )
    y = my_hubert(waveforms, labels, mask=mask, attention_mask=None)[0].mean()
    torch.testing.assert_close(x, y)



import logging
from pathlib import Path

import orjson
import torch
from filelock import FileLock
from spidr.config import MaskingConfig
from spidr.environment import set_seed, setup_environment, setup_pytorch
from spidr.tools import init_logger
from torch.utils.data import DataLoader

from .config import hubert_data_config
from .data import build_dataloader_with_labels
from .features import split_for_distributed
from .model import HuBERTPretrain

logger = logging.getLogger()


@torch.no_grad()
def validate(
    model: HuBERTPretrain,
    loader: DataLoader,
    device: torch.device,
    dtype: torch.dtype,
) -> dict[str, float]:
    model.eval()
    total_loss = torch.zeros(1, device=device)
    total_feature_loss = torch.zeros(1, device=device)
    mixed_precision = dtype != torch.float32
    for waveforms, labels, attn_mask, mask in loader:
        with torch.autocast("cuda", dtype, mixed_precision):
            loss, outputs = model(
                waveforms.to(device),
                labels.to(device),
                mask=mask.to(device),
                attention_mask=attn_mask.to(device),
            )
        total_loss += loss.mean()
        total_feature_loss += outputs["feature_loss"]
    total_loss /= len(loader)
    total_feature_loss /= len(loader)
    return {"loss": total_loss.item(), "feature_loss": total_feature_loss.item()}


def validate_all_checkpoints(
    manifest: str,
    checkpoints: str | Path,
    output: str | Path,
    *,
    seed: int = 0,
) -> None:
    init_logger()
    set_seed(seed)
    setup_pytorch(use_deterministic=False)
    setup_environment()
    device, dtype = torch.device("cuda"), torch.bfloat16
    loader = build_dataloader_with_labels(hubert_data_config(manifest), MaskingConfig())
    lock = FileLock(f"{output}.lock")
    paths = sorted(Path(checkpoints).glob("*.pt"))
    for path in split_for_distributed(paths):
        model = HuBERTPretrain.from_pretrained(path).to(device)
        losses = validate(model, loader, device, dtype)
        with lock, Path(output).open("ab") as f:
            f.write(orjson.dumps({"name": path.stem} | losses, option=orjson.OPT_APPEND_NEWLINE))


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Run validation for all existing HuBERT checkpoints")
    parser.add_argument("path_manifest", type=str, help="Path to the validation manifest file")
    parser.add_argument("path_checkpoints", type=Path, help="Directory containing all intermediate checkpoints")
    parser.add_argument("output_validation", type=Path, help="Path to the output JSONL file with validation losses")
    args = parser.parse_args()
    validate_all_checkpoints(args.path_manifest, args.path_checkpoints, args.output_validation)



import logging
from pathlib import Path
from tempfile import NamedTemporaryFile

import numpy as np
import torch
from sklearn.cluster import MiniBatchKMeans
from tqdm import tqdm

from .utils import slurm_job_tmpdir

logger = logging.getLogger()


def infer_dimension(root: str | Path) -> int:
    return torch.load(next(Path(root).rglob("*.pt")), map_location="cpu", mmap=True).size(-1)


def build_mmap_features(
    root: str | Path,
    filename: str,
    *,
    dim: int,
    seed: int,
    subsample: int,
) -> np.ndarray:
    files = sorted(Path(root).rglob("*.pt"))
    rng = np.random.default_rng(seed=seed)
    files = rng.choice(files, len(files) // subsample)
    rows = 0
    for path in tqdm(files, "Infer concatenated features shape"):
        shape = torch.load(path, map_location="cpu", mmap=True).shape
        assert len(shape) == 2 and shape[1] == dim
        rows += shape[0]
    logger.info("Shape: %s", (rows, dim))
    mmap = np.memmap(filename, dtype=np.float32, mode="w+", shape=(rows, dim))
    offset = 0
    for path in tqdm(files, "Concatenating features"):
        tensor = torch.load(path)
        length = tensor.size(0)
        mmap[offset : offset + length] = tensor.detach().numpy()
        offset += length
    return mmap


def build_kmeans(n_clusters: int, *, seed: int) -> MiniBatchKMeans:
    return MiniBatchKMeans(
        n_clusters,
        init="k-means++",
        max_iter=100,
        batch_size=10_000,
        verbose=1,
        compute_labels=False,
        random_state=seed,
        tol=0.0,
        max_no_improvement=100,
        n_init=20,
        reassignment_ratio=0.0,
    )


def fit_kmeans(
    root: str | Path,
    n_clusters: int,
    *,
    seed: int,
    subsample: int = 1,
    tmpdir: str | Path | None = None,
) -> MiniBatchKMeans:
    with NamedTemporaryFile(dir=tmpdir, suffix=".dat") as memmap_filename:
        dim = infer_dimension(root)
        features = build_mmap_features(root, memmap_filename.name, dim=dim, seed=seed, subsample=subsample)
        kmeans = build_kmeans(n_clusters, seed=seed).fit(features)
        inertia = -kmeans.score(features) / len(features)
        logger.info("Final inertia: %s", inertia)
    return kmeans


if __name__ == "__main__":
    import argparse

    import joblib

    parser = argparse.ArgumentParser(description="Train K-means on MFCC or HuBERT features")
    parser.add_argument("root_features", type=Path, help="Root directory containing feature files")
    parser.add_argument("path_kmeans", type=Path, help="Output path for the trained K-means model")
    parser.add_argument("n_clusters", type=int, help="Number of clusters")
    parser.add_argument(
        "--subsample",
        type=int,
        default=1,
        help="Subsampling factor: use every Nth file for training (default: 1 = all files).",
    )
    parser.add_argument("--seed", type=int, default=0, help="Random seed")
    parser.add_argument(
        "--tmpdir",
        type=str,
        default=slurm_job_tmpdir(),
        help="Where to store temporary concatenated features",
    )
    args = parser.parse_args()
    kmeans = fit_kmeans(
        args.root_features,
        args.n_clusters,
        seed=args.seed,
        subsample=args.subsample,
        tmpdir=args.tmpdir,
    )
    joblib.dump(kmeans, args.path_kmeans)



import math
from pathlib import Path

import torch
from spidr.config import DEFAULT_CONV_LAYER_CONFIG, DataConfig, MaskingConfig
from spidr.data.dataset import (
    BucketizeBatchSampler,
    DistributedBatchSampler,
    SpeechCollatorWithMasking,
    SpeechDataset,
    SpeechDatasetFromArchive,
    SpeechDatasetFromFiles,
    conv_length,
)
from spidr.data.masks import MaskGenerator
from torch import Tensor
from torch import distributed as dist
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader


class SpeechDatasetWithLabelsFromArchive(SpeechDatasetFromArchive):
    def __getitem__(self, index: int) -> tuple[Tensor, Tensor]:
        waveform = super().__getitem__(index)
        labels = self.manifest[index, "units"]
        return waveform, torch.tensor(labels)


class SpeechDatasetWithLabelsFromFiles(SpeechDatasetFromFiles):
    def __getitem__(self, index: int) -> tuple[Tensor, Tensor]:
        waveform = super().__getitem__(index)
        labels = self.manifest[index, "units"]
        return waveform, torch.tensor(labels)


def speech_dataset_with_labels(manifest_path: Path | str, *, normalize: bool) -> SpeechDataset:
    with Path(manifest_path).open("r", encoding="utf-8") as f:
        columns = set(f.readline().strip().split(","))
    if {"fileid", "path", "num_samples", "archive", "byte_offset", "byte_size"}.issubset(columns):
        return SpeechDatasetWithLabelsFromArchive(manifest_path, normalize=normalize)
    return SpeechDatasetWithLabelsFromFiles(manifest_path, normalize=normalize)


def crop_audio_and_labels(
    waveform: Tensor,
    label: Tensor,
    num_samples: int,
    max_sample_size: int,
    *,
    rand_crop: bool,
) -> tuple[Tensor, Tensor, int]:
    frame_offset = 0
    length = waveform.size(0)
    num_samples = min(num_samples, max_sample_size)
    if length > num_samples and rand_crop:
        frame_offset = int(torch.randint(length - num_samples, size=(1,)))
    elif length < num_samples:
        num_samples = length
    # label_offset = max(conv_length(DEFAULT_CONV_LAYER_CONFIG, frame_offset), 0)
    # num_label = conv_length(DEFAULT_CONV_LAYER_CONFIG, num_samples)
    kernel_size, stride, sample_rate = 25, 20, 16
    label_offset = max(math.floor((frame_offset - kernel_size * sample_rate) / (stride * sample_rate)) + 1, 0)
    num_label = math.floor((num_samples - kernel_size * sample_rate) / (stride * sample_rate)) + 1
    return (
        waveform[frame_offset : frame_offset + num_samples],
        label[label_offset : label_offset + num_label],
        num_samples,
    )


class SpeechWithLabelsCollatorWithMasking(SpeechCollatorWithMasking):
    def __call__(self, batch: list[tuple[Tensor, Tensor]]) -> tuple[Tensor, Tensor, Tensor, Tensor | None]:
        num_samples = max(len(wav) for wav, _ in batch) if self.enable_padding else min(len(wav) for wav, _ in batch)
        wavs_labels_len = [
            crop_audio_and_labels(wav, label, num_samples, self.max_sample_size, rand_crop=self.rand_crop)
            for wav, label in batch
        ]
        wav_list, label_list, wav_lengths = zip(*wavs_labels_len, strict=True)
        wavs = pad_sequence(wav_list, batch_first=True)
        labels = pad_sequence(label_list, batch_first=True, padding_value=-1)
        lengths = conv_length(self.conv_layer_config, torch.tensor(wav_lengths))
        batch_size, max_len = wavs.size(0), int(lengths.max())
        padding_mask = torch.arange(max_len, device=lengths.device).expand(batch_size, max_len) >= lengths[:, None]
        attn_mask = ~padding_mask[:, None, None, :].expand(batch_size, 1, max_len, max_len)
        mask_indices = self.mask_generator(padding_mask)[0]
        return wavs, labels, attn_mask, mask_indices


def build_dataloader_with_labels(data_cfg: DataConfig, mask_cfg: MaskingConfig) -> DataLoader:
    dataset = speech_dataset_with_labels(data_cfg.manifest, normalize=data_cfg.normalize)
    batch_sampler = DistributedBatchSampler(
        BucketizeBatchSampler(
            lengths=dataset.manifest["num_samples"].to_list(),
            num_buckets=data_cfg.num_buckets,
            min_len=data_cfg.min_sample_size,
            max_len=data_cfg.max_sample_size,
            max_token_count=data_cfg.max_batch_length,
            batch_size=None,
            seed=data_cfg.random_seed,
            bucket_method=data_cfg.bucket_method,
            shuffle=True,
            drop_last=data_cfg.drop_last,
        ),
        seed=data_cfg.random_seed,
        shuffle=True,
        drop_last=data_cfg.drop_last,
    )
    collate_fn = SpeechWithLabelsCollatorWithMasking(
        MaskGenerator(mask_cfg),
        max_sample_size=data_cfg.max_sample_size,
        conv_layer_config=DEFAULT_CONV_LAYER_CONFIG,
        enable_padding=data_cfg.enable_padding,
        rand_crop=data_cfg.rand_crop,
    )
    return DataLoader(
        dataset,
        batch_sampler=batch_sampler,
        num_workers=data_cfg.num_workers,
        collate_fn=collate_fn,
        pin_memory=data_cfg.pin_memory,
        prefetch_factor=data_cfg.prefetch_factor,
        persistent_workers=data_cfg.persistent_workers,
        generator=torch.Generator().manual_seed(
            data_cfg.random_seed + (dist.get_rank() if dist.is_initialized() else 0)
        ),
    )



"""Training loop."""

import logging
import os
from contextlib import ExitStack

import torch
import wandb
from spidr.checkpoint import Checkpointer
from spidr.environment import setup_training
from spidr.tools import AverageMeters, profiler_context
from spidr.train import init_wandb
from torch import GradScaler
from torch import distributed as dist
from torch.nn.parallel import DistributedDataParallel
from torch.nn.utils import clip_grad_norm_
from torch.optim import AdamW, Optimizer
from torch.optim.lr_scheduler import LRScheduler
from tqdm import tqdm

from .config import Config, read_config
from .data import build_dataloader_with_labels
from .model import HuBERTPretrain

logger = logging.getLogger()


class LinearDecayLRScheduler(LRScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        warmup_updates: int,
        max_updates: int,
    ) -> None:
        self.warmup_updates = warmup_updates
        self.max_updates = max_updates
        super().__init__(optimizer)

    def get_lr(self) -> list[float]:
        if self._step_count <= self.warmup_updates:
            return [self._step_count / self.warmup_updates * base_lr for base_lr in self.base_lrs]
        if self._step_count >= self.max_updates:
            return [0.0 for _ in self.base_lrs]
        pct_remaining = (self.max_updates - self._step_count) / (self.max_updates - self.warmup_updates)
        return [base_lr * pct_remaining for base_lr in self.base_lrs]


def train(cfg: Config) -> None:
    with ExitStack() as stack:
        logger.info("Starting job")
        setup_training(cfg.run.random_seed, use_deterministic=cfg.run.use_deterministic)
        stack.callback(dist.destroy_process_group)
        global_rank, world_size = dist.get_rank(), dist.get_world_size()
        is_main = global_rank == 0
        if is_main:
            init_wandb(cfg)
            stack.callback(wandb.finish)

        logger.info("Building model, optimizer, and dataloaders")
        device = torch.device(f"cuda:{os.environ['LOCAL_RANK']}")
        dtype = {"float32": torch.float32, "float16": torch.float16, "bfloat16": torch.bfloat16}[cfg.optimizer.dtype]
        model = HuBERTPretrain(cfg.num_classes).to(device).train()
        optimizer = AdamW(
            model.parameters(),
            lr=cfg.optimizer.lr,
            weight_decay=cfg.optimizer.weight_decay,
            betas=cfg.optimizer.betas,
            eps=cfg.optimizer.eps,
            fused=True,
        )
        scaler = GradScaler("cuda", enabled=cfg.optimizer.mixed_precision)
        scheduler = LinearDecayLRScheduler(
            optimizer,
            cfg.optimizer.warmup_steps,
            cfg.optimizer.warmup_steps + cfg.optimizer.decay_steps,
        )
        loader = build_dataloader_with_labels(cfg.data, cfg.masking)
        dist.barrier(device_ids=[device.index])
        ckpt = Checkpointer(cfg.run.dir, cfg.run.save_interval, cfg.run.keep_latest)
        ckpt.init_state(model=model, optimizer=optimizer, scheduler=scheduler, scaler=scaler)
        ckpt.load_existing_run()
        step, epoch = int(ckpt.step), int(ckpt.epoch)
        stack.callback(lambda: ckpt.save(step, epoch))
        model.compile(dynamic=True)
        ddp_model = DistributedDataParallel(model, device_ids=[device.index], find_unused_parameters=True)

        logger.info("Starting training loop")
        meters = AverageMeters(["loss", "grad_norm", "batch_size", "feature_loss"], device=device)
        profiler = stack.enter_context(profiler_context(cfg.run.dir / "trace.html" if is_main else None))
        pbar = stack.enter_context(tqdm(total=cfg.optimizer.max_steps, initial=step, disable=not is_main))
        while step < cfg.optimizer.max_steps:
            epoch += 1
            loader.batch_sampler.set_epoch(epoch)
            logger.info("Starting epoch %s", epoch)
            for waveforms, labels, attn_mask, mask in loader:
                if step >= cfg.optimizer.max_steps:
                    break
                with torch.autocast("cuda", dtype, cfg.optimizer.mixed_precision):
                    loss, outputs = ddp_model(
                        waveforms.to(device),
                        labels.to(device),
                        mask=mask.to(device),
                        attention_mask=attn_mask.to(device),
                    )
                num_frames = torch.tensor(loss.size(0), dtype=torch.long, device=device)
                dist.all_reduce(num_frames)
                loss = loss.sum() * world_size / num_frames
                scaler.scale(loss).backward()
                scaler.unscale_(optimizer)
                grad_norm = clip_grad_norm_(ddp_model.parameters(), cfg.optimizer.max_norm)
                scaler.step(optimizer)
                scaler.update()
                optimizer.zero_grad(set_to_none=True)
                lr = scheduler.get_last_lr()[0]
                scheduler.step()
                step += 1

                meters.update(
                    loss=loss.detach(),
                    batch_size=waveforms.size(0),
                    grad_norm=grad_norm,
                    feature_loss=outputs["feature_loss"],
                )
                pbar.update()
                if is_main and step % cfg.run.log_interval == 0:
                    infos = meters.pop() | {"lr": lr, "step": step, "epoch": epoch}
                    wandb.log({f"train/{key}": value for key, value in infos.items()})
                    pbar.set_postfix(loss=infos["loss"], feature_loss=infos["feature_loss"])
                    ckpt.save(step, epoch)
                profiler.step()

        if is_main:
            ckpt.save_final(step, epoch)
        logger.info("Training finished")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("config", type=str)
    cfg = read_config(parser.parse_args().config)
    train(cfg)



import logging
from pathlib import Path

import polars as pl

logger = logging.getLogger()


def find_and_symlink_best_checkpoint(checkpoint: str | Path, validation: str | Path) -> None:
    df = pl.read_ndjson(validation).sort("name")
    best = df[df["loss"].arg_min()].to_dicts()[0]
    logger.info("Best checkpoint: '%s' with %s val. loss", best["name"], best["loss"])
    (Path(checkpoint) / "best.pt").symlink_to(f"{best['name']}.pt")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("path_checkpoints", type=Path, help="Directory containing all intermediate checkpoints")
    parser.add_argument("path_validation", type=Path, help="Path to the JSONL file with validation losses")
    args = parser.parse_args()
    find_and_symlink_best_checkpoint(args.path_checkpoints, args.path_validation)



import polars as pl
from spidr.config import DEFAULT_CONV_LAYER_CONFIG
from spidr.data.dataset import conv_length


def get_fileids_from_manifest_and_units(path_manifest: str, path_units: str) -> set[str]:
    manifest = pl.scan_ndjson(path_manifest).sort("fileid").select("fileid").collect()
    fileids = set(manifest["fileid"].unique())
    units = pl.scan_ndjson(path_units).sort("fileid").select("fileid").filter(pl.col("fileid").is_in(fileids))
    assert manifest.equals(units)
    return fileids


def merge_manifest_with_units(path_manifest: str, path_units: str, *, from_mfcc: bool) -> pl.DataFrame:
    fileids = get_fileids_from_manifest_and_units(path_manifest, path_units)
    length = conv_length(DEFAULT_CONV_LAYER_CONFIG, pl.read_ndjson(path_manifest)["num_samples"].to_torch())
    return (
        pl.concat(
            (
                pl.read_ndjson(path_manifest).sort("fileid"),
                pl.scan_ndjson(path_units)
                .sort("fileid")
                .filter(pl.col("fileid").is_in(fileids))
                .drop("fileid")
                .with_columns(pl.col("units").list.gather_every(2) if from_mfcc else pl.col("units"))
                .collect(),
            ),
            how="horizontal",
        )
        .with_columns(pl.Series(name="length", values=length))
        .with_columns(pl.col("units").list.slice(offset=0, length=pl.col("length")))
        .drop("length")
    )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("path-manifest", help="Path to manifest file")
    parser.add_argument("units", help="Path to the JSONL file with units")
    parser.add_argument("output", help="Path to the output manifest file with units")
    parser.add_argument(
        "--from-mfcc",
        type=bool,
        action="store_true",
        help="Add this flag if units are derived from MFCC (10ms instead of 20ms)",
    )
    args = parser.parse_args()
    merge_manifest_with_units(args.manifest, args.units, from_mfcc=args.from_mfcc).write_ndjson(args.output)



import math
import re
from collections import defaultdict
from pathlib import Path

import torch
from spidr.config import DEFAULT_CONV_LAYER_CONFIG
from spidr.models.components import (
    ConvLayerBlock,
    ConvPositionalEmbedding,
    FeatureExtractor,
    FeatureProjection,
    FeedForward,
    SelfAttention,
    Transformer,
    TransformerLayer,
)
from torch import Tensor, nn
from torch.nn import functional as F
from torch.nn.utils.parametrizations import weight_norm

LOGIT_TEMPERATURE = 0.1
FEATURE_WEIGHT = 1.0  # Set to 1.0 instead of 10.0 to ~compensate for feature_grad_mult=0.1 in the original code


class LogitGenerator(nn.Module):
    def __init__(self, encoder_embed_dim: int, num_classes: int, final_dim: int) -> None:
        super().__init__()
        self.label_embeddings = nn.Parameter(torch.FloatTensor(num_classes, final_dim))
        nn.init.uniform_(self.label_embeddings)
        self.final_proj = nn.Linear(encoder_embed_dim, final_dim)
        self.logit_temp = nn.Buffer(torch.tensor(LOGIT_TEMPERATURE))

    def forward(self, x: Tensor, label: Tensor) -> Tensor:
        x = self.final_proj(x)
        if (label < 0).any():
            raise ValueError("Negative labels found: slicing when wrong")
        pos = torch.index_select(self.label_embeddings, 0, label).unsqueeze(0)
        negs = self.label_embeddings.unsqueeze(1).expand(-1, x.size(0), -1)
        targets = torch.cat([pos, negs], dim=0)
        logits = torch.cosine_similarity(x.float(), targets.float(), dim=-1).type_as(x) / self.logit_temp
        neg_is_pos = (pos == negs).all(-1)
        if neg_is_pos.any():
            logits[1:][neg_is_pos] = float("-inf")
        return logits.transpose(0, 1)


def hubert_base_components(
    encoder_projection_dropout: float,
    encoder_attention_dropout: float,
    encoder_ff_interm_dropout: float,
    encoder_dropout: float,
    encoder_layer_drop: float,
) -> tuple[FeatureExtractor, nn.Sequential, Transformer, nn.Parameter]:
    blocks, in_channels = nn.ModuleList(), 1
    for i, (out_channels, kernel_size, stride) in enumerate(DEFAULT_CONV_LAYER_CONFIG):
        norm = nn.GroupNorm(num_groups=out_channels, num_channels=out_channels, affine=True) if i == 0 else None
        blocks.append(ConvLayerBlock(in_channels, out_channels, kernel_size, stride, norm, bias=False))
        in_channels = out_channels
    feature_extractor = FeatureExtractor(blocks)
    pos_conv = ConvPositionalEmbedding(768, 128, 16, 1)
    pos_conv.convs[0] = weight_norm(pos_conv.convs[0], dim=2)
    pos_conv.layer_norm = nn.Identity()
    layers = nn.ModuleList()
    for _ in range(12):
        attention = SelfAttention(768, 12, qkv_bias=True, dropout=encoder_attention_dropout)
        feed_forward = FeedForward(768, 3_072, encoder_ff_interm_dropout)
        layers.append(TransformerLayer(attention, encoder_dropout, feed_forward, layer_norm_first=False))
    encoder = Transformer(layers, pos_conv, encoder_dropout, encoder_layer_drop, layer_norm_first=True)
    feature_projection = nn.Sequential(
        FeatureProjection(DEFAULT_CONV_LAYER_CONFIG[-1][0], 768),
        nn.Dropout(encoder_projection_dropout),
    )
    mask_embedding = nn.Parameter(torch.FloatTensor(768))
    return feature_extractor, feature_projection, encoder, mask_embedding


class HuBERT(nn.Module):
    def __init__(
        self,
        *,
        encoder_projection_dropout: float = 0.1,
        encoder_attention_dropout: float = 0.1,
        encoder_ff_interm_dropout: float = 0.0,
        encoder_dropout: float = 0.1,
        encoder_layer_drop: float = 0.05,
    ) -> None:
        super().__init__()
        self.feature_extractor, self.feature_projection, self.encoder, self.mask_embedding = hubert_base_components(
            encoder_projection_dropout,
            encoder_attention_dropout,
            encoder_ff_interm_dropout,
            encoder_dropout,
            encoder_layer_drop,
        )
        self.init_weights_()

    def init_weights_(self) -> None:
        nn.init.uniform_(self.mask_embedding)
        module = self.encoder.pos_conv_embed
        std = math.sqrt(4.0 / (module.embed_dim * module.kernel_size))
        for conv in module.convs:
            nn.init.normal_(conv.weight, mean=0.0, std=std)
            nn.init.constant_(conv.bias, 0.0)

    def get_intermediate_outputs(
        self,
        waveforms: Tensor,
        *,
        attention_mask: Tensor | None = None,
        num_layers: int | None = None,
        before_residual: bool = True,
    ) -> list[Tensor]:
        x = self.feature_extractor(waveforms)
        x = self.feature_projection(x)
        return self.encoder.get_intermediate_outputs(x, attention_mask, num_layers, before_residual=before_residual)

    def forward(
        self, waveforms: Tensor, *, mask: Tensor | None = None, attention_mask: Tensor | None = None
    ) -> Tensor:
        x = self.feature_extractor(waveforms)
        x = self.feature_projection(x)
        if mask is not None:
            x = torch.where(mask.unsqueeze(-1), self.mask_embedding.to(x.dtype).expand_as(x), x)
        else:
            mask = torch.ones((x.shape[0], x.shape[1]), dtype=torch.bool, device=x.device)
        return self.encoder(x, attention_mask)

    @classmethod
    def from_pretrained(cls, path: str | Path) -> "HuBERT":
        state_dict = torch.load(path, map_location="cpu")
        if "model" in state_dict:
            state_dict = state_dict["model"]
        model = cls()
        model.load_state_dict(state_dict)
        return model.eval()


class HuBERTPretrain(HuBERT):
    def __init__(
        self,
        num_classes: int,
        *,
        encoder_projection_dropout: float = 0.1,
        encoder_attention_dropout: float = 0.1,
        encoder_ff_interm_dropout: float = 0.0,
        encoder_dropout: float = 0.1,
        encoder_layer_drop: float = 0.05,
    ) -> None:
        super().__init__(
            encoder_projection_dropout=encoder_projection_dropout,
            encoder_attention_dropout=encoder_attention_dropout,
            encoder_ff_interm_dropout=encoder_ff_interm_dropout,
            encoder_dropout=encoder_dropout,
            encoder_layer_drop=encoder_layer_drop,
        )
        self.feature_weight = nn.Buffer(torch.tensor(FEATURE_WEIGHT))
        self.logit_generator = LogitGenerator(768, num_classes, 256)

    def forward(  # ty: ignore[invalid-method-override]
        self,
        waveforms: Tensor,
        labels: Tensor,
        *,
        mask: Tensor | None,
        attention_mask: Tensor | None,
    ) -> tuple[Tensor, dict[str, Tensor]]:
        x = self.feature_extractor(waveforms)
        features_pen = x.float().pow(2).mean()
        x = self.feature_projection(x)
        if mask is not None:
            x = torch.where(mask.unsqueeze(-1), self.mask_embedding.to(x.dtype).expand_as(x), x)
        else:
            mask = torch.ones((x.shape[0], x.shape[1]), dtype=torch.bool, device=x.device)
        x = self.encoder(x, attention_mask)
        mask_indices = torch.nonzero(mask, as_tuple=True)
        logits = self.logit_generator(x[mask_indices], labels[mask_indices])
        features_loss = features_pen * self.feature_weight * logits.shape[0]
        logits_loss = -F.log_softmax(logits, dim=1)[:, 0]
        return features_loss + logits_loss, {"feature_loss": features_loss, "logits_loss": logits_loss}

    @classmethod
    def from_pretrained(cls, path: str | Path) -> "HuBERTPretrain":
        state_dict = torch.load(path, map_location="cpu")
        if "model" in state_dict:
            state_dict = state_dict["model"]
        num_classes = state_dict["logit_generator.label_embeddings"].size(0)
        model = cls(num_classes)
        model.load_state_dict(state_dict)
        return model.eval()


def _fix_state_dict_key(key: str) -> str:
    if key == "masked_spec_embed":
        return "mask_embedding"
    key = re.sub(r"^wav2vec2\.", "", key)
    key = re.sub(r"^mask_generator\.", "", key)
    key = re.sub(r"^encoder\.transformer\.", "encoder.", key)
    key = re.sub(r"^feature_projection\.", "feature_projection.0.", key)
    key = re.sub(r"^encoder\.feature_projection\.", "feature_projection.0.", key)
    key = re.sub(r"\.out_proj\.", ".proj.", key)
    return re.sub(r"^encoder\.pos_conv_embed\.conv\.", "encoder.pos_conv_embed.convs.0.", key)


def state_dict_from_torchaudio_or_huggingface(state_dict: dict[str, Tensor]) -> dict[str, Tensor]:
    new_state_dict = {_fix_state_dict_key(k): v for k, v in state_dict.items()}
    qkv = {"weight": defaultdict(dict), "bias": defaultdict(dict)}
    for key, tensor in new_state_dict.items():
        match = re.match(r"^(encoder\.layers\.\d+\.attention)\.(q|k|v)_proj\.(weight|bias)$", key)
        if match:
            layer, group, param = match.groups()
            qkv[param][layer][group] = tensor
    for weight in ["weight", "bias"]:
        for layer in qkv[weight]:
            q, k, v = qkv[weight][layer]["q"], qkv[weight][layer]["k"], qkv[weight][layer]["v"]
            new_state_dict[f"{layer}.qkv.{weight}"] = torch.cat((q, k, v), dim=0)
            for group in ["q", "k", "v"]:
                del new_state_dict[f"{layer}.{group}_proj.{weight}"]
    if "logit_generator.label_embeddings" in new_state_dict:
        new_state_dict["feature_weight"] = torch.tensor(FEATURE_WEIGHT)
        new_state_dict["logit_generator.logit_temp"] = torch.tensor(LOGIT_TEMPERATURE)
    return new_state_dict


def _fix_state_dict_key_s3prl(key: str) -> str:
    if key == "mask_emb":
        return "mask_embedding"
    if key == "label_embs_concat":
        return "logit_generator.label_embeddings"
    key = re.sub(r"^wav2vec2\.", "", key)
    key = re.sub(r"^mask_generator\.", "", key)
    key = re.sub(r"^encoder\.transformer\.", "encoder.", key)
    key = re.sub(r"^feature_projection\.", "feature_projection.0.", key)
    key = re.sub(r"^encoder\.feature_projection\.", "feature_projection.0.", key)
    key = re.sub(r"\.out_proj\.", ".proj.", key)
    key = re.sub(r"\.fc1\.", ".feed_forward.intermediate_dense.", key)
    key = re.sub(r"\.fc2\.", ".feed_forward.output_dense.", key)
    key = re.sub(r"\.self_attn\.", ".attention.", key)
    key = re.sub(r"\.self_attn_layer_norm\.", ".layer_norm.", key)
    key = re.sub(r"(feature_extractor\.conv_layers\.\d+)\.0\.weight", r"\1.conv.weight", key)
    key = re.sub(r"(feature_extractor\.conv_layers\.\d+)\.2", r"\1.layer_norm", key)
    key = re.sub(r"^post_extract_proj\.", "feature_projection.0.projection.", key)
    key = re.sub(r"^layer_norm\.", "feature_projection.0.layer_norm.", key)
    key = re.sub(r"^final_proj\.", "logit_generator.final_proj.", key)
    key = re.sub(r"\.weight_g", ".parametrizations.weight.original0", key)
    key = re.sub(r"\.weight_v", ".parametrizations.weight.original1", key)
    return re.sub(r"^encoder\.pos_conv\.", "encoder.pos_conv_embed.convs.", key)


def state_dict_from_s3prl(state_dict: dict[str, Tensor]) -> dict[str, Tensor]:
    new_state_dict = {_fix_state_dict_key_s3prl(k): v for k, v in state_dict.items()}
    qkv = {"weight": defaultdict(dict), "bias": defaultdict(dict)}
    for key, tensor in new_state_dict.items():
        match = re.match(r"^(encoder\.layers\.\d+\.attention)\.(q|k|v)_proj\.(weight|bias)$", key)
        if match:
            layer, group, param = match.groups()
            qkv[param][layer][group] = tensor
    for weight in ["weight", "bias"]:
        for layer in qkv[weight]:
            q, k, v = qkv[weight][layer]["q"], qkv[weight][layer]["k"], qkv[weight][layer]["v"]
            new_state_dict[f"{layer}.qkv.{weight}"] = torch.cat((q, k, v), dim=0)
            for group in ["q", "k", "v"]:
                del new_state_dict[f"{layer}.{group}_proj.{weight}"]
    if "logit_generator.label_embeddings" in new_state_dict:
        new_state_dict["feature_weight"] = torch.tensor(FEATURE_WEIGHT)
        new_state_dict["logit_generator.logit_temp"] = torch.tensor(LOGIT_TEMPERATURE)
    return new_state_dict
